/*
 * Copyright (c) 2009 Jan Friesse <honzaf@users.sourceforge.net>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <gtk/gtk.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <stdlib.h>
#include <string.h>

#include "lwtt_parser.h"

#include "icon.xpm"

/**** CONSTANTS ****/
/*#define VERSION_MAJOR 0
#define VERSION_MINOR 1
#define VERSION_MICRO 0*/

#define XSTR(s) STR(s)
#define STR(s) #s
#define VERSION_STR XSTR(VERSION_MAJOR)"."XSTR(VERSION_MINOR)"."XSTR(VERSION_MICRO)

/** Foreground color of active text */
#define COLOR_ACTIVE_FG "Red"
/** Background color of active text */
#define COLOR_ACTIVE_BG "#fff0f0"

/** Default unit price */
#define DEFAULT_UNIT_PRICE 1.0
/** Unit price print format (printf format) */
#define UNIT_PRICE_FORMAT "%.2f"
/** Total price print format (printf format) */
#define TOTAL_PRICE_FORMAT "%.2f"

/** Auto save interval (in timer ticks -> sec) */
#define AUTO_SAVE_INTERVAL 5

#if GTK_CHECK_VERSION(2,10,0)
#define USE_GTK_STATUS_ICON 1
#endif

/**** Typedefs/enums /... ****/

/** Enum for table */
enum {
  COLUMN_TASK_NAME,
  COLUMN_TIME_CONSUMPTION,
  COLUMN_UNIT_PRICE,
  COLUMN_TOTAL_PRICE, /* Visible */
  COLUMN_UNIT_PRICE_FLOAT,  /* Not visible */
  COLUMN_TIME_CONSUMPTION_INT,
  COLUMN_ACTIVE,
  COLUMN_FG,
  COLUMN_BG,
  COLUMN_ACTIVE_NEG,
  COLUMN_TOTAL_PRICE_FLOAT,
  COLUMN_NO_OF
};

/**** STATIC VARIABLES ****/
static GtkListStore *list_items;
static GtkWidget *main_window;
static GtkWidget *execute_button;
static GtkWidget *stop_button;
static GtkWidget *properties_button;
static GtkWidget *add_button;
static GtkWidget *reset_button;
static GtkWidget *remove_button;
static GtkWidget *about_button;

#ifdef USE_GTK_STATUS_ICON
static GtkStatusIcon *status_icon;
#endif

static int time_counter;

static int running_tasks;
static gboolean main_window_maximized;
static gboolean main_window_iconified;

#ifdef USE_GTK_STATUS_ICON
static gboolean use_status_icon;
#endif

/**** Functions *****/

/*** Globally useful functions ***/

/** Load config file and save items to list_items */
static void load_config_from_app_config (GKeyFile *app_config, gboolean reshow_main_window) {
  char *task_name;
  int sect_no;
  gboolean cont;
  char group_name[256];
  float unit_price;
  GtkTreeIter iter;
  char text_float[256];
  char time_consumption_str[32];
  int time_consumption;
  int mw_x, mw_y, mw_width, mw_height;

  sect_no = 0;
  cont = TRUE;

  while (cont) {
    sprintf (group_name, "task%d", sect_no++);

    if ((task_name = g_key_file_get_string (app_config, group_name, "name", NULL)) != NULL) {
      time_consumption = 0;
      unit_price = 1;

      time_consumption = g_key_file_get_integer (app_config, group_name, "consumption", NULL);
      unit_price = (g_key_file_get_integer (app_config, group_name, "price", NULL) / 1000.0);

      sprintf (text_float, UNIT_PRICE_FORMAT, unit_price);

      gtk_list_store_append (list_items, &iter);
      gtk_list_store_set (list_items, &iter,
                    COLUMN_TASK_NAME, task_name,
                    COLUMN_UNIT_PRICE, text_float,
                    COLUMN_ACTIVE, 0,
                    COLUMN_ACTIVE_NEG, 1,
                    COLUMN_UNIT_PRICE_FLOAT, unit_price,
                    COLUMN_TIME_CONSUMPTION_INT, time_consumption,
                    -1);

      sprintf (time_consumption_str, "%d:%02d", (int)(time_consumption / 60), (int)(time_consumption % 60));
      gtk_list_store_set (list_items, &iter, COLUMN_TIME_CONSUMPTION, time_consumption_str, -1);

      sprintf (text_float, TOTAL_PRICE_FORMAT, unit_price * (time_consumption / 60.0));
      gtk_list_store_set (list_items, &iter, COLUMN_TOTAL_PRICE, text_float, COLUMN_TOTAL_PRICE_FLOAT, unit_price * (time_consumption / 60.0), -1);
      g_free (task_name);
    } else {
      cont = FALSE;
    }
  }

  mw_x = mw_y = 0;
  mw_width = mw_height = -1;

  mw_x = g_key_file_get_integer (app_config, "mainwindow", "x", NULL);
  mw_y = g_key_file_get_integer (app_config, "mainwindow", "y", NULL);
  mw_width = g_key_file_get_integer (app_config, "mainwindow", "width", NULL);
  mw_height = g_key_file_get_integer (app_config, "mainwindow", "height", NULL);
  main_window_maximized = g_key_file_get_boolean (app_config, "mainwindow", "maximized", NULL);

  gtk_window_set_default_size (GTK_WINDOW (main_window), mw_width, mw_height);

  if (reshow_main_window)
    gtk_window_reshow_with_initial_size (GTK_WINDOW (main_window));

  gtk_window_move (GTK_WINDOW (main_window), mw_x, mw_y);
  if (main_window_maximized)
    gtk_window_maximize (GTK_WINDOW (main_window));

#ifdef USE_GTK_STATUS_ICON
  use_status_icon = g_key_file_get_boolean (app_config, "statusicon", "enabled", NULL);
#endif
}

static void load_config_file (void) {
  gchar *fname;
  GKeyFile *app_config;

  app_config = g_key_file_new ();

  fname = g_build_filename (g_get_user_config_dir (), "sgtt", "sgtt.ini", NULL);
  if (g_key_file_load_from_file (app_config, fname, G_KEY_FILE_NONE, NULL)) {
    load_config_from_app_config (app_config, FALSE);
  }

  g_key_file_free (app_config);

  g_free (fname);
}

/** Save config file (save from list_items) */
static void save_config_file (void) {
  gchar *fname;
  gchar *fname_tmp;
  float unit_price;
  int time_consumption;
  GtkTreeIter iter;
  gboolean valid;
  GtkTreeModel *list_store;
  char *task_name;
  GKeyFile *app_config;
  int sect_no;
  char group_name[256];
  char *fdir;
  char *key_file_content;
  gsize key_file_len;
  FILE *f;
  char iso_time_s[256];
  time_t t;
  struct tm *tmp;
  int mw_width, mw_height, mw_x, mw_y;

  sect_no = 0;
  key_file_content = NULL;

  app_config = g_key_file_new ();

  /* Traverse list iter and store data*/
  list_store = (GtkTreeModel *)list_items;

 /* Get the first iter in the list */
  valid = gtk_tree_model_get_iter_first (list_store, &iter);
  while (valid) {
    /* Walk through the list, reading each row */
    gtk_tree_model_get (list_store, &iter,
      COLUMN_UNIT_PRICE_FLOAT, &unit_price,
      COLUMN_TIME_CONSUMPTION_INT, &time_consumption,
      COLUMN_TASK_NAME, &task_name,
      -1);

    sprintf (group_name, "task%d", sect_no++);

    g_key_file_set_string (app_config, group_name, "name", task_name);
    g_key_file_set_integer (app_config, group_name, "consumption", time_consumption);
    g_key_file_set_integer (app_config, group_name, "price", (int)(unit_price * 1000.0));

    g_free (task_name);

    valid = gtk_tree_model_iter_next (list_store, &iter);
  }

  /* Save general items */
  gtk_window_get_size (GTK_WINDOW (main_window), &mw_width, &mw_height);
  gtk_window_get_position (GTK_WINDOW (main_window), &mw_x, &mw_y);

  g_key_file_set_integer (app_config, "mainwindow", "width", mw_width);
  g_key_file_set_integer (app_config, "mainwindow", "height", mw_height);
  g_key_file_set_integer (app_config, "mainwindow", "x", mw_x);
  g_key_file_set_integer (app_config, "mainwindow", "y", mw_y);
  g_key_file_set_boolean (app_config, "mainwindow", "maximized", main_window_maximized);

#ifdef USE_GTK_STATUS_ICON
  g_key_file_set_boolean (app_config, "statusicon", "enabled", use_status_icon);
#endif

  t = time (NULL);
  tmp = gmtime (&t);
  strftime (iso_time_s, sizeof (iso_time_s), "%Y-%m-%dT%H:%M:%SZ", tmp);
  sprintf (group_name, "Simple GTK Time Tracker config file. Saved on %s", iso_time_s);

  g_key_file_set_comment (app_config, NULL, NULL, group_name, NULL);
  fname_tmp = g_build_filename (g_get_user_config_dir (), "sgtt", "sgtt.ini.new", NULL);
  fname = g_build_filename (g_get_user_config_dir (), "sgtt", "sgtt.ini", NULL);

  fdir = g_build_filename (g_get_user_config_dir (), "sgtt", NULL);
  if (g_mkdir_with_parents (fdir, 0755)) {
    fprintf (stderr, "Couldn't make config dir: %s\n", fdir);
    g_free (fdir);
    return;
  }
  g_free (fdir);

  /* Write to file */
  f = fopen (fname_tmp, "w");
  if (f) {
    key_file_content = g_key_file_to_data (app_config, &key_file_len, NULL);

    fwrite (key_file_content, 1, key_file_len, f);
    fclose (f);

    if (g_rename (fname_tmp, fname) != 0) {
      fprintf (stderr, "Couldn't save config file: %s\n", fname);
    }
  } else {
    fprintf (stderr, "Couldn't save config file: %s\n", fname_tmp);
  }

  g_key_file_free (app_config);
  g_free (key_file_content);
  g_free (fname);
  g_free (fname_tmp);
}

/*** Functions for support status icon ***/
#ifdef USE_GTK_STATUS_ICON
void status_icon_activate (GtkStatusIcon *status_icon, gpointer user_data)
{
  /* Restore main window and hide us */
  if (main_window_iconified)
    gtk_window_deiconify (GTK_WINDOW (main_window));
  else
    gtk_window_iconify (GTK_WINDOW (main_window));
}

void update_status_icon_tooltip (GtkStatusIcon *status_icon, char *title)
{
  if (status_icon == NULL) return ;

#if GTK_CHECK_VERSION(2,16,0)
  gtk_status_icon_set_tooltip_text (status_icon, title);
#else
  gtk_status_icon_set_tooltip (status_icon, title);
#endif
}

GtkStatusIcon *create_status_icon () {
  char str[256];

  status_icon = gtk_status_icon_new_from_pixbuf (gdk_pixbuf_new_from_xpm_data ((const char **)icon_xpm));

  g_signal_connect (G_OBJECT (status_icon), "activate", G_CALLBACK (status_icon_activate), NULL);

  sprintf (str, "SGTT - %d running", running_tasks);
  update_status_icon_tooltip (status_icon, str);

  return status_icon;
}
#endif

static void update_window_title (GtkWindow *window)
{
  char str[256];

  sprintf (str, "SGTT - %d running", running_tasks);

  gtk_window_set_title (window, str);
#ifdef USE_GTK_STATUS_ICON
  update_status_icon_tooltip (status_icon, str);
#endif
}

/*** Callbacks definitions ***/

/** Buttons **/
static void add_item (GtkWidget *button, gpointer data)
{
  GtkTreeIter iter;
  char text_float[256];

  sprintf (text_float, UNIT_PRICE_FORMAT, DEFAULT_UNIT_PRICE);

  gtk_list_store_append (list_items, &iter);
  gtk_list_store_set (list_items, &iter,
                      COLUMN_TASK_NAME, "Unnamed task",
                      COLUMN_UNIT_PRICE, text_float,
                      COLUMN_TIME_CONSUMPTION, "0:00",
                      COLUMN_ACTIVE, 0,
                      COLUMN_ACTIVE_NEG, 1,
                      COLUMN_UNIT_PRICE_FLOAT, DEFAULT_UNIT_PRICE,
                      COLUMN_TIME_CONSUMPTION_INT, 0,
                      -1);

  sprintf (text_float, TOTAL_PRICE_FORMAT, 0.0);
  gtk_list_store_set (list_items, &iter, COLUMN_TOTAL_PRICE, text_float, COLUMN_TOTAL_PRICE_FLOAT, 0.0, -1);
}

static void remove_item (GtkWidget *button, gpointer data)
{
  GtkTreeIter iter;
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GtkTreePath *path;
  GtkWidget *dialog;
  gboolean is_active;
  GList *items, *items_iter;
  int no_selected;
  GtkTreeRowReference  *rowref;
  GList *rowref_list;
  gint result;
  GtkTreeModel *tree_model;

  treeview = (GtkTreeView *)data;
  selection = gtk_tree_view_get_selection (treeview);

  tree_model = (GtkTreeModel *)list_items;

  items = gtk_tree_selection_get_selected_rows(selection, &tree_model);

  no_selected = 0;
  for (items_iter = items; items_iter; items_iter = g_list_next (items_iter))
    no_selected++;

  if (no_selected == 0)
    return;

  dialog = gtk_message_dialog_new (GTK_WINDOW (main_window),
    GTK_DIALOG_DESTROY_WITH_PARENT,
    GTK_MESSAGE_QUESTION,
    GTK_BUTTONS_YES_NO,
    (no_selected == 1)?"Really delete selected task?":"Really delete %d selected tasks?",
    no_selected);

  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);

  result = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);

  if (result == GTK_RESPONSE_YES) {
    rowref_list = NULL;

    for (items_iter = items; items_iter; items_iter = g_list_next (items_iter)) {
      path = (GtkTreePath *)items_iter->data;

      gtk_tree_model_get_iter (tree_model, &iter, path);
      gtk_tree_model_get (tree_model, &iter, COLUMN_ACTIVE, &is_active, -1);

      if (is_active)
        running_tasks--;

      /* We have correct number of active tasks*/

      /* Store rowref */
      rowref = gtk_tree_row_reference_new(tree_model, path);
      rowref_list = g_list_append(rowref_list, rowref);
    }

    g_list_foreach (items, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (items);

    for (items_iter = rowref_list;  items_iter;  items_iter = g_list_next (items_iter)) {
      /* Real remove */
      path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)items_iter->data);

      if (path) {
        if (gtk_tree_model_get_iter(GTK_TREE_MODEL(list_items), &iter, path))
          gtk_list_store_remove(list_items, &iter);

        gtk_tree_path_free (path);
      }
    }

    g_list_foreach(rowref_list, (GFunc) gtk_tree_row_reference_free, NULL);
    g_list_free(rowref_list);

    update_window_title (GTK_WINDOW (main_window));
  }
}

static void execute_item (GtkWidget *button, gpointer data)
{
  GtkTreeIter iter;
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GList *items, *items_iter;
  gboolean is_active;
  GtkTreeModel *tree_model;

  treeview = (GtkTreeView *)data;
  selection = gtk_tree_view_get_selection (treeview);

  tree_model = (GtkTreeModel *)list_items;

  items = gtk_tree_selection_get_selected_rows(selection, &tree_model);

  for (items_iter = items; items_iter; items_iter = g_list_next (items_iter)) {
    gtk_tree_model_get_iter (tree_model, &iter, (GtkTreePath *)items_iter->data);

    gtk_tree_model_get (tree_model, &iter, COLUMN_ACTIVE, &is_active, -1);

    if (!is_active) {
      /* Start item */
      gtk_list_store_set (list_items, &iter,
        COLUMN_ACTIVE, 1,
        COLUMN_ACTIVE_NEG, 0,
        COLUMN_BG, COLOR_ACTIVE_BG,
        COLUMN_FG, COLOR_ACTIVE_FG,
        -1);

      running_tasks++;
    }

    if (items_iter == items) {
      gtk_widget_set_sensitive (execute_button, FALSE);
      gtk_widget_set_sensitive (stop_button, TRUE);
    }
  }

  g_list_foreach (items, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (items);

  update_window_title (GTK_WINDOW (main_window));
}

static void stop_item (GtkWidget *button, gpointer data)
{
  GtkTreeIter iter;
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GList *items, *items_iter;
  gboolean is_active;
  GtkTreeModel *tree_model;

  treeview = (GtkTreeView *)data;
  selection = gtk_tree_view_get_selection (treeview);

  tree_model = (GtkTreeModel *)list_items;

  items = gtk_tree_selection_get_selected_rows(selection, &tree_model);

  for (items_iter = items; items_iter; items_iter = g_list_next (items_iter)) {
    gtk_tree_model_get_iter (tree_model, &iter, (GtkTreePath *)items_iter->data);

    gtk_tree_model_get (tree_model, &iter, COLUMN_ACTIVE, &is_active, -1);

    if (is_active) {
      /* Stop item */
      gtk_list_store_set (list_items, &iter,
        COLUMN_ACTIVE, 0,
        COLUMN_ACTIVE_NEG, 1,
        COLUMN_BG, NULL,
        COLUMN_FG, NULL,
        -1);
      running_tasks--;
    }

    if (items_iter == items) {
      gtk_widget_set_sensitive (execute_button, TRUE);
      gtk_widget_set_sensitive (stop_button, FALSE);
    }
  }
  g_list_foreach (items, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (items);

  update_window_title (GTK_WINDOW (main_window));
}

static void reset_item (GtkWidget *button, gpointer data)
{
  GtkTreeIter iter;
  GtkTreeView *treeview;
  GtkTreeSelection *selection;
  GtkWidget *dialog;
  gint result;
  char text_float[256];
  GList *items, *items_iter;
  int no_selected;
  GtkTreeModel *tree_model;

  treeview = (GtkTreeView *)data;
  selection = gtk_tree_view_get_selection (treeview);

  tree_model = (GtkTreeModel *)list_items;

  items = gtk_tree_selection_get_selected_rows(selection, &tree_model);

  no_selected = 0;
  for (items_iter = items; items_iter; items_iter = g_list_next (items_iter))
    no_selected++;

  if (no_selected == 0)
    return ;

  dialog = gtk_message_dialog_new (GTK_WINDOW (main_window),
    GTK_DIALOG_DESTROY_WITH_PARENT,
    GTK_MESSAGE_QUESTION,
    GTK_BUTTONS_YES_NO,
    (no_selected == 1)?"Really reset selected task?":"Really reset %d selected tasks?",
    no_selected);

  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);

  result = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);

  if (result == GTK_RESPONSE_YES) {
    /* Reset items */
    sprintf (text_float, TOTAL_PRICE_FORMAT, 0.0);

    for (items_iter = items; items_iter; items_iter = g_list_next (items_iter)) {
      gtk_tree_model_get_iter (tree_model, &iter, (GtkTreePath *)items_iter->data);

      gtk_list_store_set (list_items, &iter,
        COLUMN_TOTAL_PRICE, text_float,
        COLUMN_TOTAL_PRICE_FLOAT, 0.0,
        COLUMN_TIME_CONSUMPTION, "0:00",
        COLUMN_TIME_CONSUMPTION_INT, 0,
        -1);
    }
  }

}

static void load_lwtt_item (GtkWidget *button, gpointer data)
{
  GtkWidget *dialog, *msg_dialog;
  char *filename;
  GError *err = NULL;
  GKeyFile *app_config;

  dialog = gtk_file_chooser_dialog_new ("Open LWTT data file (data.xml)",
				      GTK_WINDOW (main_window),
				      GTK_FILE_CHOOSER_ACTION_OPEN,
				      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				      GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				      NULL);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));

    if (!parse_lwtt_file (filename, &app_config, &err)) {
      msg_dialog = gtk_message_dialog_new (GTK_WINDOW (main_window),
			  GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
				GTK_MESSAGE_ERROR,
				GTK_BUTTONS_OK,
				"Error during parse of file \"%s\": ", filename);
      gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (msg_dialog),
        "%s", err->message);

      gtk_dialog_run (GTK_DIALOG (msg_dialog));
      gtk_widget_destroy (msg_dialog);

      g_error_free (err);
    } else {
      gtk_list_store_clear (list_items);

      load_config_from_app_config (app_config, TRUE);

      g_key_file_free (app_config);
    }

    g_free (filename);
  }

  gtk_widget_destroy (dialog);
}

static void properties_item (GtkWidget *button, gpointer data)
{
  GtkWidget *dialog;
  GtkWidget *load_lwtt_button;
  int res;
#ifdef USE_GTK_STATUS_ICON
  GtkWidget *checkb;
#endif

  dialog = gtk_dialog_new_with_buttons ("Properties",
					GTK_WINDOW (main_window),
					GTK_DIALOG_MODAL| GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_CANCEL,
          GTK_STOCK_SAVE,
          GTK_RESPONSE_OK,
					NULL);

#ifdef USE_GTK_STATUS_ICON
  checkb = gtk_check_button_new_with_mnemonic ("Iconify to _status icon");

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkb), use_status_icon);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), checkb, FALSE, FALSE, 4);
#endif

  load_lwtt_button = gtk_button_new_with_mnemonic ("Import _LWTT config file");
  g_signal_connect (load_lwtt_button, "clicked", G_CALLBACK (load_lwtt_item), NULL);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), load_lwtt_button, FALSE, FALSE, 4);

  gtk_widget_show_all (GTK_DIALOG (dialog)->vbox);
  res = gtk_dialog_run (GTK_DIALOG (dialog));

  if (res == GTK_RESPONSE_OK) {
#ifdef USE_GTK_STATUS_ICON
    use_status_icon = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (checkb));

    if (use_status_icon && status_icon == NULL)
      create_status_icon ();

    if (!use_status_icon && status_icon != NULL) {
      g_object_unref (status_icon);
      status_icon = NULL;
    }
#endif
  }

  gtk_widget_destroy (dialog);
}

static void about_item (GtkWidget *button, gpointer data)
{
  const gchar *license =
    "Permission to use, copy, modify, and/or distribute this software for any\n"
    "purpose with or without fee is hereby granted, provided that the above\n"
    "copyright notice and this permission notice appear in all copies.\n\n"
    "THE SOFTWARE IS PROVIDED \"AS IS\" AND ISC DISCLAIMS ALL WARRANTIES\n"
    "WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n"
    "MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR\n"
    "ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n"
    "WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN\n"
    "ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF\n"
    "OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n";

  gtk_show_about_dialog (GTK_WINDOW (main_window),
			 "name", "Simple GTK Time Tracker",
			 "version", VERSION_STR,
			 "copyright", "(C) 2009 Jan Friesse <honzaf@users.sourceforge.net>",
			 "license", license,
			 "website", "http://sgtt.sourceforge.net",
			 "comments", "Program to track your activities.",
       "title", "About Simple GTK Time Tracker",
			 NULL);
}

/** Window callbacks **/

gboolean window_delete_callback (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  save_config_file ();
  return FALSE;
}

void window_destroy_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}

gboolean window_state_callback (GtkWidget *widget, GdkEventWindowState *event, gpointer user_data)
{
#ifdef USE_GTK_STATUS_ICON
  if (use_status_icon) {
    if (event->changed_mask & GDK_WINDOW_STATE_ICONIFIED) {
      main_window_iconified = event->new_window_state & GDK_WINDOW_STATE_ICONIFIED;

      if (main_window_iconified) {
        gtk_window_set_skip_taskbar_hint (GTK_WINDOW (main_window), main_window_iconified);
        gtk_widget_hide (main_window);
      } else {
        gtk_widget_show (main_window);
        gtk_window_set_skip_taskbar_hint (GTK_WINDOW (main_window), main_window_iconified);
      }
    }
  }
#endif

  if ((event->changed_mask & GDK_WINDOW_STATE_MAXIMIZED)) {
    main_window_maximized = event->new_window_state & GDK_WINDOW_STATE_MAXIMIZED;
  }
  return FALSE;
}

/** Table callbacks **/
void cell_edited_callback (GtkCellRendererText *cell, gchar *path_string, gchar *new_text, gpointer user_data)
{
  GtkTreePath *path;
  GtkTreeIter iter;
  gint column;
  float new_text_float;
  char new_text_after_conversion[256];
  char text_float[256];
  int time_consumption;
  GtkTreeModel *tree_model;

  path = gtk_tree_path_new_from_string (path_string);
  column = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (cell), "column"));

  tree_model = (GtkTreeModel *)list_items;

  gtk_tree_model_get_iter (tree_model, &iter, path);

  if (column == COLUMN_UNIT_PRICE) {
    /* Try to convert to float */
    new_text_float = atof (new_text);
    gtk_list_store_set (list_items, &iter, COLUMN_UNIT_PRICE_FLOAT, new_text_float, -1);
    sprintf (new_text_after_conversion, UNIT_PRICE_FORMAT, new_text_float);
    new_text = new_text_after_conversion;

    /* Change total price */
    gtk_tree_model_get (tree_model, &iter,
      COLUMN_TIME_CONSUMPTION_INT, &time_consumption,
      -1);
    sprintf (text_float, TOTAL_PRICE_FORMAT, new_text_float * (time_consumption / 60.0));
    gtk_list_store_set (list_items, &iter, COLUMN_TOTAL_PRICE, text_float, COLUMN_TOTAL_PRICE_FLOAT, new_text_float * (time_consumption / 60.0), -1);
  }

  gtk_list_store_set (list_items, &iter, column, new_text, -1);
}

void list_view_selection_changed_callback (GtkTreeSelection *treeselection, gpointer user_data)
{
  GtkTreeIter iter;
  GtkTreeSelection *selection;
  GList *items, *items_iter;
  gboolean is_active;
  int no_active, no_unactive;
  GtkTreeModel *tree_model;

  selection = (gtk_tree_view_get_selection ((GtkTreeView *)user_data));

  tree_model = (GtkTreeModel *)list_items;

  items = gtk_tree_selection_get_selected_rows(selection, &tree_model);

  no_active = no_unactive = 0;

  for (items_iter = items; items_iter; items_iter = g_list_next (items_iter)) {
    gtk_tree_model_get_iter (tree_model, &iter, (GtkTreePath *)items_iter->data);

    gtk_tree_model_get (tree_model, &iter, COLUMN_ACTIVE, &is_active, -1);

    if (is_active) no_active++; else no_unactive++;
  }

  g_list_foreach (items, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (items);

  gtk_widget_set_sensitive (execute_button, no_unactive > 0);
  gtk_widget_set_sensitive (stop_button, no_active > 0);

  gtk_widget_set_sensitive (remove_button, no_active + no_unactive > 0);
  gtk_widget_set_sensitive (reset_button, no_active + no_unactive > 0);
}


/** Timer callbacks **/

gboolean timer (gpointer data)
{
  gboolean valid;
  gboolean active;
  GtkTreeIter iter;
  GtkTreeModel *list_store;
  int time_consumption;
  float unit_price;
  char time_consumption_str[32];
  char total_price[256];

  list_store = (GtkTreeModel *)list_items;

 /* Get the first iter in the list */
  valid = gtk_tree_model_get_iter_first (list_store, &iter);
  while (valid) {
    /* Walk through the list, reading each row */
    gtk_tree_model_get (list_store, &iter,
      COLUMN_ACTIVE, &active,
      COLUMN_UNIT_PRICE_FLOAT, &unit_price,
      COLUMN_TIME_CONSUMPTION_INT, &time_consumption,
      -1);

    if (active) {
      time_consumption++;

      sprintf (time_consumption_str, "%d:%02d", (int)(time_consumption / 60), (int)(time_consumption % 60));
      sprintf (total_price, TOTAL_PRICE_FORMAT, unit_price * (time_consumption / 60.0));

      gtk_list_store_set (list_items, &iter,
        COLUMN_TIME_CONSUMPTION_INT, time_consumption,
        COLUMN_TIME_CONSUMPTION, time_consumption_str,
        COLUMN_TOTAL_PRICE, total_price,
        COLUMN_TOTAL_PRICE_FLOAT, unit_price * (time_consumption / 60.0),
        -1);
    }

    if ((++time_counter) % AUTO_SAVE_INTERVAL) {
      save_config_file ();
    }

    valid = gtk_tree_model_iter_next (list_store, &iter);
  }

  return TRUE;
}

/*** Functions for create window ***/

GtkWidget *list_view_create (void)
{
  GtkTreeView *list_view;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;
  GtkTreeSelection  *selection;

  list_items = gtk_list_store_new (COLUMN_NO_OF, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
    G_TYPE_FLOAT, G_TYPE_INT,
    G_TYPE_BOOLEAN, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_FLOAT);

  list_view = (GtkTreeView *)gtk_tree_view_new_with_model (GTK_TREE_MODEL (list_items));
  gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (list_view)), GTK_SELECTION_MULTIPLE);
/*  g_object_set (list_view, "enable-grid-lines", GTK_TREE_VIEW_GRID_LINES_BOTH, NULL);*/

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list_view));
  g_signal_connect(selection, "changed", (GCallback)list_view_selection_changed_callback, list_view);

  renderer = gtk_cell_renderer_text_new ();
  g_object_set(renderer, "editable-set", TRUE, NULL);
  g_signal_connect(renderer, "edited", (GCallback)cell_edited_callback, NULL);
  g_object_set_data (G_OBJECT (renderer), "column", GINT_TO_POINTER (COLUMN_TASK_NAME));
  column = gtk_tree_view_column_new_with_attributes ("Task name", renderer, "text", COLUMN_TASK_NAME,
    "editable", COLUMN_ACTIVE_NEG, NULL);
  gtk_tree_view_column_set_alignment (column, 0.5);
  gtk_tree_view_column_set_resizable  (column, TRUE);
  gtk_tree_view_column_set_sort_column_id (column, COLUMN_TASK_NAME);
  gtk_tree_view_append_column (list_view, column);
  gtk_tree_view_column_set_resizable  (column, TRUE);

  renderer = gtk_cell_renderer_text_new ();
  g_object_set(renderer, "xalign", 1.0, NULL);
  column = gtk_tree_view_column_new_with_attributes ("Time consumption [h:mm]", renderer, "text", COLUMN_TIME_CONSUMPTION,
    "foreground", COLUMN_FG, "background", COLUMN_BG, NULL);
  gtk_tree_view_column_set_alignment (column, 0.5);
  gtk_tree_view_column_set_resizable  (column, TRUE);
  gtk_tree_view_column_set_sort_column_id (column, COLUMN_TIME_CONSUMPTION_INT);
  gtk_tree_view_append_column (list_view, column);

  renderer = gtk_cell_renderer_text_new ();
  g_object_set(renderer, "editable-set", TRUE, "xalign", 1.0, NULL);
  g_signal_connect(renderer, "edited", (GCallback)cell_edited_callback, NULL);
  g_object_set_data (G_OBJECT (renderer), "column", GINT_TO_POINTER (COLUMN_UNIT_PRICE));
  column = gtk_tree_view_column_new_with_attributes ("Hour price", renderer, "text", COLUMN_UNIT_PRICE,
    "foreground", COLUMN_FG, "background", COLUMN_BG, "editable", COLUMN_ACTIVE_NEG, NULL);
  gtk_tree_view_column_set_alignment (column, 0.5);
  gtk_tree_view_column_set_resizable  (column, TRUE);
  gtk_tree_view_column_set_sort_column_id (column, COLUMN_UNIT_PRICE_FLOAT);
  gtk_tree_view_append_column (list_view, column);

  renderer = gtk_cell_renderer_text_new ();
  g_object_set(renderer, "xalign", 1.0, NULL);
  column = gtk_tree_view_column_new_with_attributes ("Total price", renderer, "text", COLUMN_TOTAL_PRICE,
    "foreground", COLUMN_FG, "background", COLUMN_BG, NULL);
  gtk_tree_view_column_set_alignment (column, 0.5);
  gtk_tree_view_column_set_resizable  (column, TRUE);
  gtk_tree_view_column_set_sort_column_id (column, COLUMN_TOTAL_PRICE_FLOAT);
  gtk_tree_view_append_column (list_view, column);

  return (GtkWidget *)list_view;
}

GtkWidget *toolbar_create (GtkWidget *treeview)
{
  GtkWidget *hbox;
  GtkWidget *button;

  hbox = gtk_hbutton_box_new ();

  button = gtk_button_new_from_stock (GTK_STOCK_EXECUTE);
  g_signal_connect (button, "clicked", G_CALLBACK (execute_item), treeview);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  execute_button = button;
  gtk_widget_set_sensitive (button, FALSE);

  button = gtk_button_new_from_stock (GTK_STOCK_STOP);
  g_signal_connect (button, "clicked", G_CALLBACK (stop_item), treeview);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  stop_button = button;
  gtk_widget_set_sensitive (button, FALSE);

  button = gtk_button_new_from_stock (GTK_STOCK_ADD);
  g_signal_connect (button, "clicked", G_CALLBACK (add_item), NULL);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  add_button = button;

  button = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
  g_signal_connect (button, "clicked", G_CALLBACK (remove_item), treeview);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  gtk_widget_set_sensitive (button, FALSE);
  remove_button = button;

  button = gtk_button_new_from_stock (GTK_STOCK_CLEAR);
  g_signal_connect (button, "clicked", G_CALLBACK (reset_item), treeview);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  gtk_widget_set_sensitive (button, FALSE);
  reset_button = button;

  button = gtk_button_new_from_stock (GTK_STOCK_PREFERENCES);
  g_signal_connect (button, "clicked", G_CALLBACK (properties_item), NULL);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  properties_button = button;

  button = gtk_button_new_from_stock (GTK_STOCK_ABOUT);
  g_signal_connect (button, "clicked", G_CALLBACK (about_item), NULL);
  gtk_container_add (GTK_CONTAINER (hbox), button);
  about_button = button;

  return hbox;
}

GtkWidget *main_window_create (void)
{
  GtkWidget *window;
  GtkWidget *scroll_win;
  GtkWidget *list_view;
  GtkWidget *toolbar;
  GtkWidget *vbox = gtk_vbox_new (FALSE, 5);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (window_delete_callback), NULL);
  g_signal_connect (G_OBJECT (window), "destroy", G_CALLBACK (window_destroy_callback), NULL);
  g_signal_connect (G_OBJECT (window), "window_state_event", G_CALLBACK (window_state_callback), NULL);

  gtk_window_set_icon (GTK_WINDOW (window), gdk_pixbuf_new_from_xpm_data ((const char **)icon_xpm));

  scroll_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll_win), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  list_view = list_view_create ();

  toolbar = toolbar_create (list_view);

  gtk_box_pack_start (GTK_BOX (vbox), toolbar, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), scroll_win, TRUE, TRUE, 0);

  gtk_container_add (GTK_CONTAINER (scroll_win), list_view);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_window_set_default_size (GTK_WINDOW (window), 500, 200);

  main_window = window;

  return window;
}

/*** MAIN ***/
int main(int argc, char *argv[])
{
  GError *error;
  gboolean start_iconified;
  GOptionEntry entries[] = {{"iconify", 'i', 0, G_OPTION_ARG_NONE, &start_iconified, "Start aplication iconified", NULL}, {NULL, }};

  start_iconified = FALSE;
  error = NULL;
  time_counter = 0;
  running_tasks = 0;
  main_window_maximized = FALSE;
  main_window_iconified = FALSE;

#ifdef USE_GTK_STATUS_ICON
  use_status_icon = FALSE;
  status_icon = NULL;
#endif

  if (!gtk_init_with_args (&argc, &argv, NULL, entries, NULL, &error)) {
    fprintf (stderr, "Error: %s\n", error->message);
    g_error_free (error);

    return 1;
  }

  main_window_create ();

  load_config_file ();

  g_timeout_add (60 * 1000, timer, NULL);

  if (start_iconified) {
    gtk_window_iconify (GTK_WINDOW (main_window));
  }

  gtk_widget_show_all (main_window);

  update_window_title (GTK_WINDOW (main_window));

#ifdef USE_GTK_STATUS_ICON
  if (use_status_icon) {
    create_status_icon ();
  }
#endif

  gtk_main ();

  return 0;
}
